#include "lagarith.h"
#include "prediction.cpp"
#include "compact.cpp"
#include "threading.cpp"

#ifndef X64_BUILD
#include "huffyuv_a.h"
#include "convert_yuy2.cpp"
#include "convert_yv12.cpp"
#else
#include "convert.h"
#endif

#pragma warning(disable:4731)

#define swap(x,y) { unsigned int xchng = *(unsigned int*)(&x); x=y; *(unsigned int*)(&y)=xchng;}

// initalize the codec for compression
DWORD CodecInst::CompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut){
	if ( started  == 0x1337 ){
		CompressEnd();
	}
	started = 0;

	nullframes = GetPrivateProfileInt("settings", "nullframes", false, "lagarith.ini")>0;
	multithreading = GetPrivateProfileInt("settings", "multithreading", false, "lagarith.ini")>0;

	if ( int error = CompressQuery(lpbiIn,lpbiOut) != ICERR_OK ){
		return error;
	}

	width = lpbiIn->biWidth;
	height = lpbiIn->biHeight;

	format = lpbiIn->biBitCount;
	length = width*height*format/8;

	unsigned int buffer_size;
	if ( format < RGB24 ){
		buffer_size = align_round(width,32)*height*format/8+1024;
	} else {
		buffer_size = align_round(width,16)*height*4+1024;
	}
	
	buffer = (unsigned char *)lag_aligned_malloc(buffer, buffer_size,16,"buffer");
	buffer2 = (unsigned char *)lag_aligned_malloc(buffer2, buffer_size,16,"buffer2");
	prev = (unsigned char *)lag_aligned_malloc(prev, buffer_size,16,"prev");

	if ( !buffer || !buffer2 || !prev ){
		lag_aligned_free(buffer,"buffer");
		lag_aligned_free(buffer2,"buffer2");
		lag_aligned_free(prev,"prev");
		return ICERR_MEMORY;
	}

	int cb_size=width*height*5/4;

	if ( !cObj.InitCompressBuffers( cb_size ) ){
		return ICERR_MEMORY;
	}

	if ( multithreading ){
		int code = InitThreads( true );
		if ( code != ICERR_OK ){
			return code;
		}
	}
	
	started = 0x1337;

	return ICERR_OK;
}

// get the maximum size a compressed frame will take;
// 105% of image size + 4KB should be plenty even for random static

DWORD CodecInst::CompressGetSize(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut){
	return (DWORD)( align_round(lpbiIn->biWidth,16)*lpbiIn->biHeight*lpbiIn->biBitCount/8*1.05 + 4096);
}

// release resources when compression is done

DWORD CodecInst::CompressEnd(){
	if ( started  == 0x1337 ){
		if ( multithreading ){
			EndThreads();
		}

		lag_aligned_free( buffer,"buffer");
		lag_aligned_free( buffer2,"buffer2");
		lag_aligned_free( prev,"prev");
//		lag_aligned_free( lossy_buffer,"lossy_buffer");
		cObj.FreeCompressBuffers();
	}
	started = 0;
	return ICERR_OK;
}

// encode a typical RGB24 frame
int CodecInst::CompressRGB24(ICCOMPRESS* icinfo){

	if ( format == RGB24 ){
		if ( (width%4)==0 && SolidFrameTestRGB24(icinfo) ) return ICERR_OK;
	} else {
		if ( SolidFrameTestRGB32(icinfo) ) return ICERR_OK;
	}

	unsigned int stride= align_round(width,16); // round up if needed for alignment
	const unsigned char * src=in;
	const unsigned int block_len=stride*height;
	const unsigned int pixels = width*height;
	unsigned char * b=buffer;
	unsigned char * g=buffer+block_len;
	unsigned char * r=buffer+block_len*2;

	// convert the interleaved channels to planer, aligning if needed
	if ( format == RGB24 ){
		if ( stride==width){
			for ( unsigned int a=0;a<width*height*3;a+=3){
				*b++=src[a]-src[a+1];
				*g++=src[a+1];
				*r++=src[a+2]-src[a+1];
			}
		} else if ( width%4 == 0 ) { // if the width is not mod 4, there will be padding 
			for ( unsigned int y=0;y<height;y++){
				unsigned int x=0;
				for (;x<width;x++){
					b[y*stride+x]=src[0]-src[1];
					g[y*stride+x]=src[1];
					r[y*stride+x]=src[2]-src[1];
					src+=3;
				}
				unsigned char i,j,k;
				i=b[y*stride+x-1];
				j=g[y*stride+x-1];
				k=r[y*stride+x-1];
				for ( ;x<stride;x++){
					b[y*stride+x]=i;
					g[y*stride+x]=j;
					r[y*stride+x]=k;
				}
			}
		} else {
			const unsigned int src_stride=align_round(width*3,4);
			for ( unsigned int y=0;y<height;y++){
				unsigned int x=0;
				for (;x<width;x++){
					unsigned int pos = y*src_stride+x*3;
					b[y*stride+x]=src[pos+0]-src[pos+1];
					g[y*stride+x]=src[pos+1];
					r[y*stride+x]=src[pos+2]-src[pos+1];
				}
				unsigned char i,j,k;
				i=b[y*stride+x-1];
				j=g[y*stride+x-1];
				k=r[y*stride+x-1];
				for ( ;x<stride;x++){
					b[y*stride+x]=i;
					g[y*stride+x]=j;
					r[y*stride+x]=k;
				}
			}
		}
	// RGB32 format
	} else {
		if ( stride==width ){
			for ( unsigned int a=0;a<width*height*4;a+=4){
				*b++=src[a]-src[a+1];
				*g++=src[a+1];
				*r++=src[a+2]-src[a+1];
			}
		} else {
			for ( unsigned int y=0;y<height;y++){
				unsigned int x=0;
				for (;x<width;x++){
					int pos = y*width*4+x*4;
					b[y*stride+x]=src[pos+0]-src[pos+1];
					g[y*stride+x]=src[pos+1];
					r[y*stride+x]=src[pos+2]-src[pos+1];
				}
				
				unsigned char i,j,k;
				i=b[y*stride+x-1];
				j=g[y*stride+x-1];
				k=r[y*stride+x-1];
				for ( ;x<stride;x++){
					b[y*stride+x]=i;
					g[y*stride+x]=j;
					r[y*stride+x]=k;
				}
			}
		}
	}
	// buffer now has aligned planar data

	int size=0;
	if ( !multithreading ) { // perform prediction
		
		unsigned char * dest;
		if ( width!=stride ){
			dest=(unsigned char *)align_round(out,16); // make sure dest is aligned (since it's allocated by calling program)
		} else {
			dest=buffer2;
		}

#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_BlockPredict(buffer,dest,stride,block_len,1);
			SSE2_BlockPredict(buffer+block_len,dest+block_len,stride,block_len,1);
			SSE2_BlockPredict(buffer+block_len*2,dest+block_len*2,stride,block_len,1);
		} else {
			MMX_BlockPredict(buffer,dest,stride,block_len,SSE,1);
			MMX_BlockPredict(buffer+block_len,dest+block_len,stride,block_len,SSE,1);
			MMX_BlockPredict(buffer+block_len*2,dest+block_len*2,stride,block_len,SSE,1);
		}
#else 
		SSE2_BlockPredict(buffer,dest,stride,block_len,1);
		SSE2_BlockPredict(buffer+block_len,dest+block_len,stride,block_len,1);
		SSE2_BlockPredict(buffer+block_len*2,dest+block_len*2,stride,block_len,1);
#endif

		// remove alignment padding if needed
		if ( width!= stride ){
			unsigned char * padded = dest;
			unsigned char * stripped = buffer2;
			for ( unsigned int y=0;y<height*3;y++){
				memcpy(stripped+y*width,padded+y*stride,width);
			}
		}


		size = cObj.compact(buffer2,out+9,pixels); 
		size+=9;

		*(UINT32*)(out+1)=size;
		size+= cObj.compact(buffer2+pixels,out+size, pixels);

		*(UINT32*)(out+5)=size;
		size+= cObj.compact(buffer2+pixels*2,out+size, pixels);

	} else { 
		info_a.source=buffer;
		info_a.dest=out+9;
		info_a.length=pixels;
		lag_ResumeThread(info_a.thread);
		info_b.source=buffer+block_len;
		info_b.dest=prev;
		info_b.length=pixels;
		lag_ResumeThread(info_b.thread);

#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_BlockPredict(buffer+block_len*2,buffer2,stride,block_len,1);
		} else {
			MMX_BlockPredict(buffer+block_len*2,buffer2,stride,block_len,SSE,1);
		}
#else
		SSE2_BlockPredict(buffer+block_len*2,buffer2,stride,block_len,1);
#endif

		unsigned char * stripped=buffer2;
		if ( width!= stride ){
			unsigned char * padded = buffer2;
			stripped += block_len;
			for ( unsigned int y=0;y<height*3;y++){
				memcpy(stripped+y*width,padded+y*stride,width);
			}
		}

		size=cObj.compact(stripped,prev+length/2,pixels);
		while ( info_a.length || info_b.length){
			Sleep(0);
		}

		//assert(out[9] < 4 || out[9]==255);
		//assert(prev[0] < 4 || prev[0]==255);
		//assert((prev+length/2)[0] < 4 || (prev+length/2)[0]==255);

		int sizea = info_a.size;
		int sizeb = info_b.size;
		*(UINT32*)(out+1)=sizea+9;
		*(UINT32*)(out+5)=sizeb+9+sizea;
		memcpy(out+sizea+9,prev,sizeb);
		memcpy(out+sizea+sizeb+9,prev+length/2,size);
		size+=sizea+sizeb+9;
	}

	if ( (width%4) == 0 ){
		out[0]=ARITH_RGB24;
	} else {
		out[0]=UNALIGNED_RGB24; // old versions didn't handle the padding correctly,
								// this tag lets the codec handle the versions differently 
	}

	icinfo->lpbiOutput->biSizeImage = size;
	

	return ICERR_OK;
}

// Test to see if the frame is a solid color, and encode it as a 'byte' or
// 'pixel' frame if so. A byte frame results from a frame that is a solid
// greyscale color. A pixel frame results from a frame that is any other 
// solid color. All other frames are compressed using the other compression
// routines. While the compression gains from using byte/pixel frames are
// usually fairly small, these frames are much faster to decode.

bool CodecInst::SolidFrameTestRGB24(ICCOMPRESS* icinfo ){
	if ( width & 3 || memcmp(in,in+48,length-48) )
		return false;

	for (int a =3; a < 48; a+=3){
		if (in[a] != in[0] || in[a+1]!=in[1] || in[a+2]!=in[2] ){
			return false;
		}
	}

	if ( in[0]==in[1] && in[1] == in[2] ){
		icinfo->lpbiOutput->biSizeImage = 2;
		out[0] = BYTEFRAME;
		out[1] = in[0];
	} else {
		icinfo->lpbiOutput->biSizeImage = 4;
		out[0] = PIXELFRAME;
		out[1] = in[0];
		out[2] = in[1];
		out[3] = in[2];
	}
	return true;
}

// See if the frame can be compressed as a null frame or RGBA pixel frame

bool CodecInst::SolidFrameTestRGB32(ICCOMPRESS* icinfo){
	if ( memcmp(buffer,in+16,width*height*4-16))
		return false;

	if (( sizeof(int)==4 &&( *(UINT32*)(in) == *(UINT32*)(in+4) && *(UINT32*)(in+4) == *(UINT32*)(in+8) && *(UINT32*)(in+8) == *(UINT32*)(in+12) ))
		|| sizeof(int)==8 && ( *(unsigned int *)(in) == *(unsigned int *)(in+8) ) ){
		*(UINT32*)(out+1)=*(UINT32*)in;
		if ( use_alpha ){
			out[0]=PIXELFRAME_ALPHA;
			icinfo->lpbiOutput->biSizeImage =5;
		} else {
			if ( in[0]==in[1]&&in[0]==in[2] ){
				out[0]=BYTEFRAME;
				icinfo->lpbiOutput->biSizeImage=2;
			} else {
				out[0]=PIXELFRAME;
				icinfo->lpbiOutput->biSizeImage=4;
			}
		}
		return true;
	}
	return false;
}

// compress an RGBA frame (when alpha compression is enabled)

int CodecInst::CompressRGBA(ICCOMPRESS* icinfo){

#ifndef X64_BUILD
	const unsigned int mod = (SSE2?16:8);
#else
	const unsigned int mod = 16;
#endif

	const unsigned int stride= align_round(width,mod); // round up if needed for alignment
	const unsigned char * src=in;
	const unsigned int block_len=stride*height;
	const unsigned int pixels = width*height;
	unsigned char * b=buffer;
	unsigned char * g=buffer+block_len;
	unsigned char * r=buffer+block_len*2;
	unsigned char * alpha=buffer+block_len*3;
	
	// convert the interleaved channels to planer, aligning if needed
	if ( stride==width){
		for ( unsigned int a=0;a<width*height*4;a+=4){
			*b++=src[a]-src[a+1];
			*g++=src[a+1];
			*r++=src[a+2]-src[a+1];
			*alpha++=src[a+3];
		}
	} else {
		for ( unsigned int y=0;y<height;y++){
			for (unsigned int x=0;x<width;x++){
				b[y*stride+x]=src[0]-src[1];
				g[y*stride+x]=src[1];
				r[y*stride+x]=src[2]-src[1];
				alpha[y*stride+x]=src[3];
				src+=4;
			}
			for ( ;x<stride;x++){
				b[y*stride+x]=src[-4]-src[-3];
				g[y*stride+x]=src[-3];
				r[y*stride+x]=src[-2]-src[-3];
				alpha[y*stride+x]=src[-1];
			}
		}
	}
	
	int size;
	if ( !multithreading ){

		// perform prediction
		unsigned char * dest;
		if ( width!=stride ){
			dest=(unsigned char *)align_round(out,16);
		} else {
			dest =buffer2;
		}

#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_BlockPredict(buffer,dest,stride,block_len,1);
			SSE2_BlockPredict(buffer+block_len,dest+block_len,stride,block_len,1);
			SSE2_BlockPredict(buffer+block_len*2,dest+block_len*2,stride,block_len,1);
			SSE2_BlockPredict(buffer+block_len*3,dest+block_len*3,stride,block_len,1);
		} else {
			MMX_BlockPredict(buffer,dest,stride,block_len,SSE,1);
			MMX_BlockPredict(buffer+block_len,dest+block_len,stride,block_len,SSE,1);
			MMX_BlockPredict(buffer+block_len*2,dest+block_len*2,stride,block_len,SSE,1);
			MMX_BlockPredict(buffer+block_len*3,dest+block_len*3,stride,block_len,SSE,1);
		}
#else 
		SSE2_BlockPredict(buffer,dest,stride,block_len,1);
		SSE2_BlockPredict(buffer+block_len,dest+block_len,stride,block_len,1);
		SSE2_BlockPredict(buffer+block_len*2,dest+block_len*2,stride,block_len,1);
		SSE2_BlockPredict(buffer+block_len*3,dest+block_len*3,stride,block_len,1);
#endif

		// remove alignment padding if needed
		if ( width!= stride ){
			unsigned char * padded = dest;
			unsigned char * stripped = buffer2;
			for ( unsigned int y=0;y<height*4;y++){
				memcpy(stripped,padded+stride*y,width);
				stripped+=width;
			}
		}

		out[0]=ARITH_ALPHA;
		size = 13;
		size += cObj.compact(buffer2,out+size,pixels);
		*(int*)(out+1)=size;
		size += cObj.compact(buffer2+pixels,out+size, pixels);
		*(int*)(out+5)=size;
		size += cObj.compact(buffer2+pixels*2,out+size, pixels);
		*(int*)(out+9)=size;
		size += cObj.compact(buffer2+pixels*3,out+size, pixels);
	} else {
		info_a.source=buffer;
		info_a.dest=out+13;
		info_a.length=pixels;
		lag_ResumeThread(info_a.thread);

		info_b.source=buffer+block_len;
		info_b.dest=prev;
		info_b.length=pixels;
		lag_ResumeThread(info_b.thread);

		info_c.source=buffer+block_len*2;
		info_c.dest=prev+align_round(pixels*3/2,16);
		info_c.length=pixels;
		lag_ResumeThread(info_c.thread);

#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_BlockPredict(buffer+block_len*3,buffer2,stride,block_len,1);
		} else {
			MMX_BlockPredict(buffer+block_len*3,buffer2,stride,block_len,SSE,1);
		}
#else
		SSE2_BlockPredict(buffer+block_len*3,buffer2,stride,block_len,1);
#endif

		if ( width!= stride ){
			unsigned char * padded = buffer2;
			unsigned char * stripped = buffer2+block_len;
			for ( unsigned int y=0;y<height;y++){
				memcpy(stripped,padded+stride*y,width);
				stripped+=width;
			}
			size=cObj.compact(buffer2+block_len,prev+align_round(pixels*3,16),pixels);
		} else {
			size=cObj.compact(buffer2,prev+align_round(pixels*3,16),pixels);
		}

		while ( info_a.length || info_b.length || info_c.length){
			Sleep(0);
		}
		out[0]=ARITH_ALPHA;
		int sizea = info_a.size;
		int sizeb = info_b.size;
		int sizec = info_c.size;
		*(UINT32*)(out+1)=sizea+13;
		*(UINT32*)(out+5)=sizeb+13+sizea;
		*(UINT32*)(out+9)=sizeb+13+sizea+sizec;
		memcpy(out+sizea+13,prev,sizeb);
		memcpy(out+sizea+sizeb+13,prev+align_round(pixels*3/2,16),sizec);
		memcpy(out+sizea+sizeb+sizec+13,prev+align_round(pixels*3,16),size);
		size+=sizea+sizeb+sizec+13;
	}

	icinfo->lpbiOutput->biSizeImage = size;
	return ICERR_OK;
}

// compress a YUY2 keyframe
int CodecInst::CompressYUY2(ICCOMPRESS* icinfo){

#ifndef X64_BUILD
	const unsigned int mod		= (SSE2?15:7);
#else
	const unsigned int mod		= 15;
#endif
	const unsigned int stride	= width*2;
	const unsigned int amod		= mod+1-((width/2)&mod);
	const unsigned int ay_stride= align_round(width,mod+1);
	const unsigned int ac_stride= align_round(width/2,mod+1);
	const unsigned int ay_len	= ay_stride*height;
	const unsigned int ac_len	= ac_stride*height;
	unsigned char * ysrc		= buffer;
	unsigned char * usrc		= ysrc+ay_len;
	unsigned char * vsrc		= usrc+ac_len;
	unsigned char * ydest		= buffer2;
	unsigned char * udest		= ydest+ay_len;
	unsigned char * vdest		= udest+ac_len;

	if ( !((width/2) & mod)){
		if ( icinfo->lpbiInput->biCompression != FOURCC_UYVY ){
			for ( unsigned int a=0;a<width*height/2;a++){
				ysrc[a*2]	= in[a*4+0];
				usrc[a]		= in[a*4+1];
				ysrc[a*2+1]	= in[a*4+2];
				vsrc[a]		= in[a*4+3];
			}
		} else { // UYVY format
			for ( unsigned int a=0;a<width*height/2;a++){
				usrc[a]		= in[a*4+0];
				ysrc[a*2]	= in[a*4+1];
				vsrc[a]		= in[a*4+2];
				ysrc[a*2+1]	= in[a*4+3];
			}
		}
	} else {
		const unsigned char * src = in;
		unsigned char * u=usrc;
		unsigned char * v=vsrc;
		unsigned char * y=ysrc;
		if ( icinfo->lpbiInput->biCompression != FOURCC_UYVY ){
			for ( unsigned int h=0;h<height;h++){
				unsigned int x=0;
				for (;x<width*2;x+=4){
					y[0]=src[0];
					u[0]=src[1];
					y[1]=src[2];
					v[0]=src[3];
					y+=2;
					u++;
					v++;
					src+=4;
				}
				unsigned char i,j,k;
				i=y[-1];
				j=u[-1];
				k=v[-1];
				for ( unsigned int z=x/4;z<ac_stride;z++){
					u[0]=j;
					v[0]=k;
					u++;
					v++;
				}
				for ( x/=2;x<ay_stride;x+=2){
					y[0]=i;
					y[1]=i;
					y+=2;
				}
			}
		} else { // UYVY format
			for ( unsigned int h=0;h<height;h++){
				unsigned int x=0;
				for (;x<width*2;x+=4){
					u[0]=src[0];
					y[0]=src[1];
					v[0]=src[2];
					y[1]=src[3];
					y+=2;
					u++;
					v++;
					src+=4;
				}
				unsigned char i,j,k;
				i=y[-1];
				j=u[-1];
				k=v[-1];
				for ( unsigned int z=x/4;z<ac_stride;z++){
					u[0]=j;
					v[0]=k;
					u++;
					v++;
				}
				for ( x/=2;x<ay_stride;x+=2){
					y[0]=i;
					y[1]=i;
					y+=2;
				}
			}
		}

	}

	int size;
	if ( !multithreading ){

#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_Predict_YUY2(ysrc,ydest,ay_stride,height,1);
			SSE2_Predict_YUY2(usrc,udest,ac_stride,height,0);
			SSE2_Predict_YUY2(vsrc,vdest,ac_stride,height,0);
		} else {
			MMX_Predict_YUY2(ysrc,ydest,ay_stride,height,SSE,1);
			MMX_Predict_YUY2(usrc,udest,ac_stride,height,SSE,0);
			MMX_Predict_YUY2(vsrc,vdest,ac_stride,height,SSE,0);
		}
#else
		SSE2_Predict_YUY2(ysrc,ydest,ay_stride,height,1);
		SSE2_Predict_YUY2(usrc,udest,ac_stride,height,0);
		SSE2_Predict_YUY2(vsrc,vdest,ac_stride,height,0);
#endif
		swap(ysrc,ydest);
		swap(usrc,udest);
		swap(vsrc,vdest);

		// remove alignment padding
		if ( width&mod ){// remove luminance padding 
			for ( unsigned int y=0;y< height;y++){
				memcpy(ydest+y*width,ysrc+y*ay_stride,width);
			}
			swap(ysrc,ydest);
		}
		if ( (width/2)&mod ){ // remove chroma padding
			unsigned int y;
			for ( y=0;y< height;y++){
				memcpy(udest+y*width/2,usrc+y*ac_stride,width/2);
			}
			for ( y=0;y< height;y++){
				memcpy(vdest+y*width/2,vsrc+y*ac_stride,width/2);
			}
			swap(usrc,udest);
			swap(vsrc,vdest);
		}


		size = cObj.compact(ysrc,out+9,width*height)+9;
		*(UINT32*)(out+1)=size;
		size+=cObj.compact(usrc,out+size,width*height/2);
		*(UINT32*)(out+5)=size;
		size+=cObj.compact(vsrc,out+size,width*height/2);
	} else { 
		int pixels = width*height;
		info_a.source=ysrc;
		info_a.dest=out+9;
		info_a.length=pixels;
		lag_ResumeThread(info_a.thread);
		info_b.source=usrc;
		info_b.dest=buffer2;
		info_b.length=pixels/2;
		lag_ResumeThread(info_b.thread);

#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_Predict_YUY2(vsrc,vdest,ac_stride,height,0);
		} else {
			MMX_Predict_YUY2(vsrc,vdest,ac_stride,height,SSE,0);
		}
#else
		SSE2_Predict_YUY2(vsrc,vdest,ac_stride,height,0);
#endif
		swap(vsrc,vdest);
		
		if ( (width/2)&mod ){ // remove chroma padding
			for ( unsigned int y=0;y< height;y++){
				memcpy(vdest+y*width/2,vsrc+y*ac_stride,width/2);
			}
			swap(vsrc,vdest);
		}

		size=cObj.compact(vsrc,prev,pixels/2);
		while ( info_a.length || info_b.length){
			Sleep(0);
		}

		assert( *(__int64*)info_a.dest != 0 );

		int sizea = info_a.size;
		int sizeb = info_b.size;
		*(UINT32*)(out+1)=sizea+9;
		*(UINT32*)(out+5)=sizeb+9+sizea;
		memcpy(out+sizea+9,buffer2,sizeb);
		memcpy(out+sizea+sizeb+9,prev,size);
		size+=sizea+sizeb+9;
	}
	out[0] = ARITH_YUY2;
	icinfo->lpbiOutput->biSizeImage = size;
	//assert( *(__int64*)(out+9) != 0 );
	assert( *(__int64*)(out+*(UINT32*)(out+1)) != 0 );
	assert( *(__int64*)(out+*(UINT32*)(out+5)) != 0 );
	return ICERR_OK;
}

// Compress a YV12 keyframe
int CodecInst::CompressYV12(ICCOMPRESS* icinfo){

#ifdef X64_BUILD
	const unsigned int mod = 15;
#else
	const unsigned int mod = (SSE2?16:8) - 1;
#endif

	const unsigned int hw = width/2;
	const unsigned int hh = height/2;

	const unsigned int y_len	= width*height;
	const unsigned int c_len	= y_len/4;
	const unsigned int yu_len	= y_len+c_len;
	const unsigned int len		= yu_len+c_len;

	const unsigned int y_stride	= align_round(width,mod+1);
	const unsigned int c_stride	= align_round(width/2,mod+1);

	const unsigned int ay_len	= y_stride*height;
	const unsigned int ac_len	= c_stride*height/2;
	const unsigned int ayu_len	= ay_len+ac_len;

	const unsigned char * ysrc;
	const unsigned char * usrc;
	const unsigned char * vsrc;

	const unsigned int in_aligned = !((int)in&mod);

	//note: chroma has only half the width of luminance, it may need to be aligned seperately

	if ( ((width/2)&mod) == 0 ) { // no alignment padding needed
		if ( in_aligned ){ // no alignment needed
			ysrc = in;
			usrc = in+y_len;
			vsrc = in+yu_len;
		} else {  // data is naturally aligned,	input buffer is not
			memcpy(buffer,in,len);
			ysrc = buffer;
			usrc = buffer+y_len;
			vsrc = buffer+yu_len;
		}
	} else if ( (width&mod) == 0 ){ // chroma needs alignment padding
		if ( in_aligned ){
			ysrc = in;
		} else {
			memcpy(buffer,in,y_len);
			ysrc = buffer;
		}

		buffer += y_len; // step over luminance (not always needed...)

		for ( unsigned int y=0;y<hh*2;y++){
			memcpy(buffer+y*c_stride,in+y*hw+y_len,hw);
			unsigned char val=buffer[y*c_stride+hw-1];
			for ( unsigned int x=hw;x<c_stride;x++){
				buffer[y*c_stride+x]=val;
			}
		}
		usrc = buffer;
		vsrc = buffer+ac_len;

		buffer -= y_len;
	} else { // both chroma and luminance need alignment padding

		// align luminance
		unsigned int y;
		for ( y=0;y<height;y++){
			memcpy(buffer+y*y_stride,in+y*width,width);
			unsigned char val=buffer[y*y_stride+width-1];
			for ( unsigned int x=width;x<y_stride;x++){
				buffer[y*y_stride+x]=val;
			}
		}
		ysrc = buffer;
		buffer += ay_len;

		for ( y=0;y<hh*2;y++){
			memcpy(buffer+y*c_stride,in+y*hw+y_len,hw);
			unsigned char val=buffer[y*c_stride+hw-1];
			for ( unsigned int x=hw;x<c_stride;x++){
				buffer[y*c_stride+x]=val;
			}
		}
		usrc = buffer;
		vsrc = buffer+c_stride*hh;

		buffer -= ay_len;
	}
	// done aligning, aligned data is in 'in' (which may actually be prev) or buffer 

	int size;
	if ( !multithreading ){

		unsigned char *buffer3=(unsigned char *)align_round(out,16);

		//set up dest buffers based on if alignment padding needs removal later
		unsigned char * ydest;
		unsigned char * udest;
		unsigned char * vdest;
		if ( width&mod ){
			ydest=buffer3;
		} else {
			ydest=buffer2;
		}
		if ( (width/2)&mod ){
			udest=buffer3+ay_len;
			vdest=buffer3+ay_len+ac_len;
		} else {
			udest=buffer2+ay_len;
			vdest=buffer2+ay_len+ac_len;
		}

		// perform prediction
#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_BlockPredict(ysrc,ydest,y_stride,ay_len,0);
			SSE2_BlockPredict(usrc,udest,c_stride,ac_len,0);
			SSE2_BlockPredict(vsrc,vdest,c_stride,ac_len,0);
		} else {
			MMX_BlockPredict(ysrc,ydest,y_stride,ay_len,SSE,0);
			MMX_BlockPredict(usrc,udest,c_stride,ac_len,SSE,0);
			MMX_BlockPredict(vsrc,vdest,c_stride,ac_len,SSE,0);
		}
#else
		SSE2_BlockPredict(ysrc,ydest,y_stride,ay_len,0);
		SSE2_BlockPredict(usrc,udest,c_stride,ac_len,0);
		SSE2_BlockPredict(vsrc,vdest,c_stride,ac_len,0);
#endif

		// remove alignment padding, all data will end up in buffer2
		if ( width&mod ){// remove luminance padding 
			for ( unsigned int y=0;y< height;y++){
				memcpy(buffer2+y*width,ydest+y*y_stride,width);
			}
		}
		if ( hw&mod ){ // remove chroma padding
			for ( unsigned int y=0;y< hh*2;y++){
				memcpy(buffer2+y_len+y*hw,udest+y*c_stride,hw);
			}
		}

		size = 9;
		size+=cObj.compact(buffer2,out+size,y_len);
		*(UINT32*)(out+1)=size;
		size+=cObj.compact(buffer2+y_len,out+size,c_len);
		*(UINT32*)(out+5)=size;
		size+=cObj.compact(buffer2+yu_len,out+size,c_len);
	} else {
		info_a.source=ysrc;
		info_a.dest=out+9;
		info_a.length=y_len;
		lag_ResumeThread(info_a.thread);
		info_b.source=usrc;
		info_b.dest=buffer2;
		info_b.length=c_len;
		lag_ResumeThread(info_b.thread);

		unsigned char *vdest=buffer2+ayu_len;

#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_BlockPredict(vsrc,vdest,c_stride,ac_len,0);
		} else {
			MMX_BlockPredict(vsrc,vdest,c_stride,ac_len,SSE,0);
		}
#else
		SSE2_BlockPredict(vsrc,vdest,c_stride,ac_len,0);
#endif

		// remove alignment if needed
		if ( hw&mod ){ // remove chroma padding
			for ( unsigned int y=0;y< hh;y++){
				memcpy((unsigned char*)(int)vsrc+y*hw,vdest+y*c_stride,hw);
			}
		} else {
			unsigned char * temp=(unsigned char*)(int)vsrc;
			vsrc = vdest;
			vdest = temp;
		}

		size=cObj.compact(vsrc,vdest,c_len);
		while ( info_a.length || info_b.length){
			Sleep(0);
		}
		int sizea = info_a.size;
		int sizeb = info_b.size;
		*(UINT32*)(out+1)=sizea+9;
		*(UINT32*)(out+5)=sizeb+9+sizea;
		memcpy(out+sizea+9,buffer2,sizeb);
		memcpy(out+sizea+sizeb+9,vdest,size);
		size+=sizea+sizeb+9;
	}
	out[0] = ARITH_YV12;
	icinfo->lpbiOutput->biSizeImage = size;
	return ICERR_OK;
}

int CodecInst::CompressReduced(ICCOMPRESS *icinfo){

#ifndef X64_BUILD
	const unsigned int mod = (SSE2?16:8);
#else
	const unsigned int mod = 16;
#endif
	const unsigned int ry_stride = align_round(width/2,mod);
	const unsigned int rc_stride = align_round(width/4,mod);
	const unsigned int ry_size   = ry_stride*height/2;
	const unsigned int rc_size   = rc_stride*height/4;
	const unsigned int ryu_size  = ry_size+rc_size;
	const unsigned int y_size    = width*height;
	const unsigned int c_size    = width*height/4;
	const unsigned int yu_size   = y_size+c_size;
	const unsigned int ry_bytes  = width*height/4;
	const unsigned int rc_bytes  = width*height/4/4;

	unsigned char * ysrc = buffer;
	unsigned char * usrc = buffer+ry_size;
	unsigned char * vsrc = buffer+ryu_size;

	//crash();

#ifndef X64_BUILD
	reduce_res(in,ysrc,buffer2,width,height,SSE2,SSE,MMX);
	reduce_res(in+y_size,usrc,buffer2,width/2,height/2,SSE2,SSE,MMX);
	reduce_res(in+yu_size,vsrc,buffer2,width/2,height/2,SSE2,SSE,MMX);
#else
	reduce_res(in,ysrc,buffer2,width,height,1,0,0);
	reduce_res(in+y_size,usrc,buffer2,width/2,height/2,1,0,0);
	reduce_res(in+yu_size,vsrc,buffer2,width/2,height/2,1,0,0);
#endif

	unsigned int size;

	if ( !multithreading ){
		unsigned char * ydest = buffer2;
		unsigned char * udest = buffer2+ry_size;
		unsigned char * vdest = buffer2+ryu_size;

		// perform prediction
#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_BlockPredict(ysrc,ydest,ry_stride,ry_size,0);
			SSE2_BlockPredict(usrc,udest,rc_stride,rc_size,0);
			SSE2_BlockPredict(vsrc,vdest,rc_stride,rc_size,0);
		} else {
			MMX_BlockPredict(ysrc,ydest,ry_stride,ry_size,SSE,0);
			MMX_BlockPredict(usrc,udest,rc_stride,rc_size,SSE,0);
			MMX_BlockPredict(vsrc,vdest,rc_stride,rc_size,SSE,0);
		}
#else
		SSE2_BlockPredict(ysrc,ydest,ry_stride,ry_size,0);
		SSE2_BlockPredict(usrc,udest,rc_stride,rc_size,0);
		SSE2_BlockPredict(vsrc,vdest,rc_stride,rc_size,0);
#endif

		if ( (width/2)%mod ){// remove luminance padding
			for ( unsigned int y=0;y< height/2;y++){
				memcpy(ysrc+y*width/2,ydest+y*ry_stride,width/2);
			}
		} else {
			ysrc=ydest;
		}
		if ( rc_stride != width/4 ){ // remove chroma padding
			unsigned int y;
			for ( y=0;y<height/4;y++){
				memcpy(usrc+y*width/4,udest+y*rc_stride,width/4);
			}
			for ( y=0;y<height/4;y++){
				memcpy(vsrc+y*width/4,vdest+y*rc_stride,width/4);
			}
		} else {
			usrc=udest;
			vsrc=vdest;
		}

		size = 9;
		size+=cObj.compact(ysrc,out+size,ry_bytes);
		*(UINT32*)(out+1)=size;
		size+=cObj.compact(usrc,out+size,rc_bytes);
		*(UINT32*)(out+5)=size;
		size+=cObj.compact(vsrc,out+size,rc_bytes);
	} else {
		info_a.source=ysrc;
		info_a.dest=out+9;
		info_a.length=ry_bytes;
		lag_ResumeThread(info_a.thread);
		info_b.source=usrc;
		info_b.dest=prev;
		info_b.length=rc_bytes;
		lag_ResumeThread(info_b.thread);

		unsigned char *vdest = buffer2;

#ifndef X64_BUILD
		if ( SSE2 ){
			SSE2_BlockPredict(vsrc,vdest,rc_stride,rc_size,0);
		} else {
			MMX_BlockPredict(vsrc,vdest,rc_stride,rc_size,SSE,0);
		}
#else
		SSE2_BlockPredict(vsrc,vdest,rc_stride,rc_size,0);
#endif

		// remove alignment if needed
		if ( rc_stride != width/4 ){ // remove chroma padding
			vsrc=vdest;
			vdest=buffer2+rc_size;
			for ( unsigned int y=0;y< height/4;y++){
				memcpy(vdest+y*width/4,vsrc+y*rc_stride,width/4);
			}
		}
		vsrc = vdest;
		vdest = buffer2+rc_size*2;

		size=cObj.compact(vsrc,vdest,rc_bytes);
		while ( info_a.length || info_b.length){
			Sleep(0);
		}
		int sizea = info_a.size;
		int sizeb = info_b.size;
		*(UINT32*)(out+1)=sizea+9;
		*(UINT32*)(out+5)=sizeb+9+sizea;
		memcpy(out+sizea+9,prev,sizeb);
		memcpy(out+sizea+sizeb+9,vdest,size);
		size+=sizea+sizeb+9;
	}
	out[0] = REDUCED_RES;
	icinfo->lpbiOutput->biSizeImage = size;
	return ICERR_OK;

	return ICERR_OK;
}

// This downsamples the colorspace if the set encoding colorspace is lower than the
// colorspace of the input video

int CodecInst::CompressLossy(ICCOMPRESS * icinfo ){

	unsigned char * lossy_buffer=prev;

	const unsigned char * const stored_in=in;
	const unsigned char * src;
	DWORD ret_val;
	if ( format >= RGB24 ){
		unsigned char * dest;
		if (lossy_option==1 ){
			dest=lossy_buffer;
		} else {
			dest=buffer;
		}

#ifndef X64_BUILD
		if ( format == RGB24 ){
			mmx_ConvertRGB24toYUY2(in,dest,width*3,width*2,width,height);
		} else if ( format == RGB32 ){
			mmx_ConvertRGB32toYUY2((const unsigned int *)in,(unsigned int *)dest,width,width/2,width,height);
		}
#else
		if ( format == RGB24 ){
			for ( unsigned int a=0;a<width*height;a+=4){
				buffer2[a*4+0 ]=in[a*3+0];
				buffer2[a*4+1 ]=in[a*3+1];
				buffer2[a*4+2 ]=in[a*3+2];
				buffer2[a*4+3 ]=0;
				buffer2[a*4+4 ]=in[a*3+3];
				buffer2[a*4+5 ]=in[a*3+4];
				buffer2[a*4+6 ]=in[a*3+5];
				buffer2[a*4+7 ]=0;
				buffer2[a*4+8 ]=in[a*3+6];
				buffer2[a*4+9 ]=in[a*3+7];
				buffer2[a*4+10]=in[a*3+8];
				buffer2[a*4+11]=0;
				buffer2[a*4+12]=in[a*3+9];
				buffer2[a*4+13]=in[a*3+10];
				buffer2[a*4+14]=in[a*3+11];
				buffer2[a*4+15]=0;
			}
			mmx_ConvertRGB32toYUY2(buffer2,dest,width,width/2,width,height);
		} else if ( format == RGB32 ){
			mmx_ConvertRGB32toYUY2(in,dest,width,width/2,width,height);
		}
#endif
		if ( lossy_option==1){
			in=dest;
			ret_val = CompressYUY2(icinfo);
		} else {
			src=dest;
		}
	} else {
		src=in;
	}

	if ( lossy_option > 1 ) {
		unsigned char * dst2=lossy_buffer;

#ifndef X64_BUILD
		if ( SSE ){
			isse_yuy2_to_yv12(src,width*2,width*2,dst2,dst2+width*height+width*height/4,dst2+width*height,width,width/2,height);
		} else {
			mmx_yuy2_to_yv12( src,width*2,width*2,dst2,dst2+width*height+width*height/4,dst2+width*height,width,width/2,height);
		}
#else
		isse_yuy2_to_yv12(src,width*2,width*2,dst2,dst2+width*height+width*height/4,dst2+width*height,width,width/2,height);
#endif
		in=lossy_buffer;
		if ( lossy_option == 2 ){
			ret_val = CompressYV12(icinfo);
		} else {
			ret_val = CompressReduced(icinfo);
		}
	}
	in=stored_in;
	return ret_val;
}

// called to compress a frame; the actual compression will be
// handed off to other functions depending on the color space and settings

DWORD CodecInst::Compress(ICCOMPRESS* icinfo, DWORD dwSize) {

//	crash();

	out = (unsigned char *)icinfo->lpOutput;
	in  = (unsigned char *)icinfo->lpInput;

	if ( icinfo->lFrameNum == 0 ){
		if ( started != 0x1337 ){
			if ( int error = CompressBegin(icinfo->lpbiInput,icinfo->lpbiOutput) != ICERR_OK )
				return error;
		}
	} else if ( nullframes ){
		// compare in two parts, video is more likely to change in middle than at bottom
		unsigned int pos = length/2+15;
		pos&=(~15);
		if (!memcmp(in+pos,prev+pos,length-pos) && !memcmp(in,prev,pos) ){
			icinfo->lpbiOutput->biSizeImage =0;
			*icinfo->lpdwFlags = 0;
			return ICERR_OK;
		}
	}
	if ( icinfo->lpckid ){
		*icinfo->lpckid = 'cd';
	}
	
	int ret_val;


	if ( lossy_option ){
		if ( format == YV12 && lossy_option == 3 ){
			ret_val = CompressReduced(icinfo);
		} else {
			ret_val = CompressLossy(icinfo);
		}
	} else if ( format == RGB24 || ( format == RGB32 && !use_alpha) ){
		ret_val = CompressRGB24(icinfo);

	/*#ifdef LOSSLESS_CHECK
			unsigned char * compare = (unsigned char *)malloc(width*height*4+1024);
			const unsigned char * const holdin = in;
			unsigned char * holdout = out;
			in = out;
			out = compare;
			ArithDecompress();
			for ( int a=0;a<width*height*format/8;a++){
				if ( holdin[a] != compare[a] ){
					__asm int 3;
				}
			}

			free(compare);
			in = holdin;
			out = holdout;
	#endif*/

	} else if ( format == YUY2 ){
		ret_val = CompressYUY2(icinfo);
	} else if ( format == YV12 ){
		ret_val = CompressYV12(icinfo);
	} else {
		ret_val = CompressRGBA(icinfo);
	}

	/*if ( interlaced ){
		out[0]|=INTERLACED_FLAG;
	}*/

	if ( nullframes ){
		memcpy( prev,in, length);
	}

	if (ret_val == ICERR_OK ){
		*icinfo->lpdwFlags = AVIIF_KEYFRAME;
	}

	return (DWORD)ret_val;
}

#include "decompression.cpp"

//MessageBox (HWND_DESKTOP, msg, "Error", MB_OK | MB_ICONEXCLAMATION);