#include "compact_old.cpp"

// initalize the codec for decompression
DWORD CodecInst::DecompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut){
	if ( started == 0x1337){
		DecompressEnd();
	}
	started = 0;

	if ( int error = DecompressQuery(lpbiIn,lpbiOut) != ICERR_OK ){
		return error;
	}

	int buffer_size;

	width = lpbiIn->biWidth;
	height = lpbiIn->biHeight;

#ifndef X64_BUILD
	detectFlags(&SSE2, &SSE, &MMX);
#endif

	format = lpbiOut->biBitCount;
	
	length = width*height*format/8;

	buffer_size = length+2048;
	if ( format >= RGB24 ){
		buffer_size=width*height*4+2048;
	}

	buffer = (unsigned char *)lag_aligned_malloc( buffer,buffer_size,16,"buffer");
	buffer2 = (unsigned char *)lag_aligned_malloc( prev,buffer_size,16,"prev");

	if ( !buffer || !buffer2 ){
		return ICERR_MEMORY;
	}

	if ( !cObj.InitCompressBuffers( width*height*5/4 ) ){
		return ICERR_MEMORY;
	}

	multithreading = GetPrivateProfileInt("settings", "multithreading", false, "lagarith.ini")>0;
	if ( multithreading ){
		int code = InitThreads( false);
		if ( code != ICERR_OK ){
			return code;
		}
	}
	started = 0x1337;
	return ICERR_OK;
}

// release resources when decompression is done
DWORD CodecInst::DecompressEnd(){
	if ( started == 0x1337 ){
		if ( multithreading ){
			EndThreads();
		}

		lag_aligned_free( buffer,"buffer");
		lag_aligned_free( buffer2, "buffer2");
		cObj.FreeCompressBuffers();

	}
	started=0;
	return ICERR_OK;
}

// decompress a typical RGB24 frame
void CodecInst::ArithDecompress(){

//	crash();

	const unsigned int pixels = width*height;
	const unsigned int stride32 = align_round(width*4,8);
	const unsigned int stride24 = align_round(width*3,4);
	int skip = *(UINT32*)(in+1);

	unsigned char *planar = NULL;

	if ( multithreading && in[0] != OBSOLETE_ARITH){
		planar = buffer2;
	} else {
		if ( width%2 || format == RGB24 || (width%4 && in[0] != UNALIGNED_RGB24)  ){
			planar = out;
		} else {
			planar = buffer;
		}
	}

	if ( in[0] != OBSOLETE_ARITH ){
		uncompact_macro(in+9,planar,pixels,width,height,&info_a,RGB24);
		uncompact_macro(in+skip,planar+pixels,pixels,width,height,&info_b,RGB24);
		skip= *(UINT32*)(in+5);
		uncompact_macro(in+skip,planar+pixels*2,pixels,width,height,NULL,RGB24);

		if ( multithreading ){
			ASM_BlockRestore(planar+pixels*2,width,width*height,1);
		}

		wait_for_threads(2);
		
	} else {
		cObj.old_uncompact(in+9,planar,pixels);
		cObj.old_uncompact(in+skip,planar+pixels,pixels);
		skip= *(UINT32*)(in+5);
		cObj.old_uncompact(in+skip,planar+pixels*2,pixels);
	}

	unsigned char * interleaved;
	if ( out == planar ){
		interleaved = buffer;
	} else {
		interleaved = out;
	}

	if ( multithreading && in[0] != OBSOLETE_ARITH){
		//crash();
		unsigned char * b=planar;
		unsigned char * g=b+pixels;
		unsigned char * r=g+pixels;
		if ( width%4==0 || in[0] == UNALIGNED_RGB24){
			if ( format == RGB24 ){
				for ( unsigned int y=0;y<height;y++){
					for ( unsigned int x=0;x<width;x++){
						out[y*stride24+x*3+0]=b[y*width+x]+g[y*width+x];
						out[y*stride24+x*3+1]=g[y*width+x];
						out[y*stride24+x*3+2]=r[y*width+x]+g[y*width+x];
					}
				}
			} else {
				for ( unsigned int x=0;x<pixels;x++){
					out[x*4+0]=b[x]+g[x];
					out[x*4+1]=g[x];
					out[x*4+2]=r[x]+g[x];
					out[x*4+3]=255;
				}
			} 
		} else {	
			for ( unsigned int x=0; x<pixels; x++){
				unsigned int pos24=x*3;
				buffer[pos24+0]=b[x]+g[x];
				buffer[pos24+1]=g[x];
				buffer[pos24+2]=r[x]+g[x];
			}
			// replace the part that was cut off with the scan line below it
			unsigned int wpos=width*height*3;
			unsigned int cut_len=stride24*height-wpos;
			unsigned int rpos=wpos-stride24;
			memcpy(buffer+wpos,buffer+rpos,cut_len);
			if ( format == RGB24 ){
				memcpy(out,buffer,stride24*height);
			} else {
				for ( unsigned int y=0;y<height;y++){
					for ( unsigned int x=0;x<width;x++){
						out[y*width*4+x*4+0]=buffer[y*stride24+x*3+0];
						out[y*width*4+x*4+1]=buffer[y*stride24+x*3+1];
						out[y*width*4+x*4+2]=buffer[y*stride24+x*3+2];
						out[y*width*4+x*4+3]=255;
					}
				}
			}
		}
		return;
	}

	if ( width % 2 ){ // odd width, need to align scanlines for prediction restore

		//crash();

		for ( unsigned int y=0;y<height;y++){
			for ( unsigned int x=0;x<width;x++){
				unsigned int posi = y*stride32+x*4;
				unsigned int posp = y*width+x;
				interleaved[posi+0]=planar[posp];
				interleaved[posi+1]=planar[posp+pixels];
				interleaved[posi+2]=planar[posp+pixels*2];
				interleaved[posi+3]=0;
			}
			
			unsigned int posi = y*stride32+width*4;

			interleaved[posi+0]=0;
			interleaved[posi+1]=0;
			interleaved[posi+2]=0;
			interleaved[posi+3]=0;
		
		}
	} else { // no alignment needed
		for ( unsigned int a=0;a<pixels;a++){
			interleaved[a*4+0]=planar[a];
			interleaved[a*4+1]=planar[a+pixels];
			interleaved[a*4+2]=planar[a+pixels*2];
			interleaved[a*4+3]=0;
		}
	}
	interleaved[3]=255; // set alpha to fully opaque, Restoration will fill in the rest

	// undo pixel prediction
#ifndef X64_BUILD
	MMX_Restore32(interleaved,stride32,stride32*height,SSE,1);
#else
	Restore32(interleaved,stride32,stride32*height);
#endif

	if ( format == RGB24 ){
		if ( (width%4)==0 || in[0] == UNALIGNED_RGB24 ){ // normal 32->24 conversion
			for ( unsigned int y=0; y<height; y++){
				for ( unsigned int x=0; x<width; x++){
					unsigned int pos32=y*stride32+x*4;
					unsigned int pos24=y*stride24+x*3;
					out[pos24+0]=interleaved[pos32+0];
					out[pos24+1]=interleaved[pos32+1];
					out[pos24+2]=interleaved[pos32+2];
				}
			}
		} else { // this is an old unaligned frame
			for ( unsigned int y=0; y<height; y++){
				for ( unsigned int x=0; x<width; x++){
					unsigned int pos32=y*stride32+x*4;
					unsigned int pos24=y*width*3+x*3;
					out[pos24+0]=interleaved[pos32+0];
					out[pos24+1]=interleaved[pos32+1];
					out[pos24+2]=interleaved[pos32+2];
				}
			}
			// replace the part that was cut off with the scan line below it
			unsigned int wpos=width*height*3;
			unsigned int cut_len=stride24*height-wpos;
			unsigned int rpos=wpos-stride24;
			memcpy(out+wpos,out+rpos,cut_len);
		}
	} else if ( width%2 && in[0] == UNALIGNED_RGB24 ){ // padded 32 -> stripped 32
		for ( unsigned int y=0; y<height; y++){
			memcpy( out+width*4*y,interleaved+stride32*y,width*4);
		}
	} else if ( width%4 && in[0] != UNALIGNED_RGB24 ){ // this is an old unaligned frame

		unsigned int y;
		for ( y=0; y<height;y++){
			for ( unsigned int x=0; x<width;x++){
				buffer2[y*width*3+x*3+0]=interleaved[y*stride32+x*4+0];
				buffer2[y*width*3+x*3+1]=interleaved[y*stride32+x*4+1];
				buffer2[y*width*3+x*3+2]=interleaved[y*stride32+x*4+2];
			}
		}

		for ( y=0; y<height;y++){
			for ( unsigned int x=0; x<width;x++){
				out[y*width*4+x*4+0]=buffer2[y*stride24+x*3+0];
				out[y*width*4+x*4+1]=buffer2[y*stride24+x*3+1];
				out[y*width*4+x*4+2]=buffer2[y*stride24+x*3+2];
			}
		}
		
		// replace the part that was cut off with the scan line below it
		unsigned int cut_pixels=(stride24*height-width*height*3)/3;
		unsigned int wpos=width*height*4-cut_pixels*4;
		unsigned int rpos=wpos-width*4;
		memcpy(out+wpos,out+rpos,cut_pixels*4);
	}
}

// decompress a RGB24 byte frame

void CodecInst::ByteDecompress(){
	memset(out,in[1],align_round(width*format/8,4)*height);
}

// decompress a RGB24 pixel frame
void CodecInst::PixelDecompress(){
	unsigned int r,g,b;
	b = in[1];
	g = in[2];
	r = in[3];
	if ( format == RGB24 ){
		for ( unsigned int a =0; a < length; a+=3){
			out[a]=b;
			out[a+1]=g;
			out[a+2]=r;
		}
	} else {
		unsigned int color=b + (g<<8) + (r<<16)+0xff000000;
		if( sizeof(unsigned int)==8 ){
			color|=color<<32;
		}
		for ( unsigned int a =0; a < length; a+=sizeof(unsigned int)){
			*(unsigned int*)(out+a)=color;
		}	
	}
}

void CodecInst::AlphaPixelDecompress(){
	if ( format == RGB24 ){
		unsigned int r,g,b;
		b = in[1];
		g = in[2];
		r = in[3];
		for ( unsigned int x=0; x<length;x+=3){
			out[x]=b;
			out[x+1]=g;
			out[x+2]=r;
		}
	} else if ( format == RGB32 ){
		unsigned int rgba = *(UINT32*)(in+1);
		if ( sizeof(unsigned int)==8 ){
			rgba|=rgba<<32;
		}
		for ( unsigned int x=0; x<length;x+=sizeof(unsigned int)){
			*(unsigned int*)(out+x)=rgba;
		}
	}
}

// decompress a RGBA keyframe
void CodecInst::ArithRGBADecompress(){

	unsigned char * planar;
	if ( width % 2 && !multithreading ){
		planar = out;
	} else {
		planar = buffer;
	}

	unsigned int pixels = width*height;
	int skip = *(UINT32*)(in+1);
	uncompact_macro(in+13,planar,pixels,width,height,&info_a,RGB32);
	uncompact_macro(in+skip,planar+pixels,pixels,width,height,&info_b,RGB32);
	if ( format == RGB32 ){
		skip= *(UINT32*)(in+9);
		uncompact_macro(in+skip,planar+pixels*3,pixels,width,height,&info_c,RGB32);
	}
	skip= *(UINT32*)(in+5);
	uncompact_macro(in+skip,planar+pixels*2,pixels,width,height,NULL,RGB32);

	if ( multithreading ){
		ASM_BlockRestore(planar+pixels*2,width,pixels,1);
	}

	wait_for_threads(3);

	const unsigned int stride32=align_round(width*4,8);
	const unsigned int stride24=align_round(width*3,4);

	if ( multithreading ){
		//crash();
		unsigned char * b=planar;
		unsigned char * g=b+pixels;
		unsigned char * r=g+pixels;
		unsigned char * a=r+pixels;
	
		if ( format == RGB24 ){
			for ( unsigned int y=0;y<height;y++){
				for ( unsigned int x=0;x<width;x++){
					out[y*stride24+x*3+0]=b[y*width+x]+g[y*width+x];
					out[y*stride24+x*3+1]=g[y*width+x];
					out[y*stride24+x*3+2]=r[y*width+x]+g[y*width+x];
				}
			}
		} else {
			for ( unsigned int x=0;x<pixels;x++){
				out[x*4+0]=b[x]+g[x];
				out[x*4+1]=g[x];
				out[x*4+2]=r[x]+g[x];
				out[x*4+3]=a[x];
			}
		}
		return;
	}


	unsigned char * interleaved;

	if ( width % 2 ){
		interleaved = buffer;
	} else {
		interleaved = out;
	}

	if ( format == RGB32 ){
		for ( unsigned int y=0; y<height;y++){
			unsigned int x;
			for ( x=0; x<width;x++){
				interleaved[y*stride32+x*4+0]=planar[y*width+x];
				interleaved[y*stride32+x*4+1]=planar[y*width+x+pixels];
				interleaved[y*stride32+x*4+2]=planar[y*width+x+pixels*2];
				interleaved[y*stride32+x*4+3]=planar[y*width+x+pixels*3];
			}
			if ( y < height-1 ){
				interleaved[y*stride32+x*4+0]=0;
				interleaved[y*stride32+x*4+1]=0;
				interleaved[y*stride32+x*4+2]=0;
				interleaved[y*stride32+x*4+3]=0;
			}
		}
	} else {
		for ( unsigned int y=0; y<height;y++){
			unsigned int x;
			for ( x=0; x<width;x++){
				interleaved[y*stride32+x*4+0]=planar[y*width+x];
				interleaved[y*stride32+x*4+1]=planar[y*width+x+pixels];
				interleaved[y*stride32+x*4+2]=planar[y*width+x+pixels*2];
				interleaved[y*stride32+x*4+3]=0;
			}
			if ( y < height-1 ){
				interleaved[y*stride32+x*4+0]=0;
				interleaved[y*stride32+x*4+1]=0;
				interleaved[y*stride32+x*4+2]=0;
				interleaved[y*stride32+x*4+3]=0;
			}
		}
	}

	// undo pixel prediction
#ifndef X64_BUILD
	MMX_Restore32(interleaved,stride32,stride32*height,SSE,1);
#else
	Restore32(interleaved,stride32,stride32*height);
#endif

	if ( format == RGB24 ){
		const unsigned int stride24=align_round(width*3,4);
		for ( unsigned int y=0; y<height; y++){
			for ( unsigned int x=0; x<width; x++){
				unsigned int pos32=y*stride32+x*4;
				unsigned int pos24=y*stride24+x*3;
				out[pos24+0]=interleaved[pos32+0];
				out[pos24+1]=interleaved[pos32+1];
				out[pos24+2]=interleaved[pos32+2];
			}
		}
	} else if ( width%2 ){ // padded 32 -> stripped 32
		for ( unsigned int y=0; y<height; y++){
			memcpy( out+width*4*y,interleaved+stride32*y,width*4);
		}
	}
}

// Decompress a YUY2 keyframe
void CodecInst::ArithYUY2Decompress(){
	unsigned char * dst = out;
	unsigned char * dst2 = buffer;
	if ( format == YUY2 ){
		dst = buffer;
		dst2= out;
	}

	int size=*(UINT32*)(in+1);
	uncompact_macro(in+9,dst,width*height,width,height,&info_a,YUY2);
	// special case: RLE detected a solid Y value (compressed size = 2),
	// need to set 2nd Y value for restoration to work right
	if ( size == 11 ){ 
		dst[1]=dst[0];
	}
	uncompact_macro(in+size,dst+width*height,width*height/2,width/2,height,&info_b,YUY2);
	size = *(UINT32*)(in+5);
	uncompact_macro(in+size,dst+width*height+width*height/2,width*height/2,width/2,height,NULL,YUY2);

	wait_for_threads(2);

	unsigned char *y,*u,*v;
	y=dst;
	u=y+width*height;
	v=u+width*height/2;

	for ( unsigned int a=0;a<width*height*2;a+=4){
		dst2[a]=*y;
		dst2[a+1]=*u;
		dst2[a+2]=*(y+1);
		dst2[a+3]=*v;
		y+=2;
		u++;
		v++;
	}

#ifndef X64_BUILD
	asm_MedianRestore(dst2,dst2+width*height*2, width*2);
	if ( format != YUY2 ){
		if ( format == RGB24 ){
			mmx_YUY2toRGB24(buffer,out,buffer+width*height*2,width*2);
		} else {
			mmx_YUY2toRGB32(buffer,out,buffer+width*height*2,width*2);
		} 
	}
#else
	MedianRestoreYUY2(dst2,dst2+width*height*2, width*2);

	if ( format != YUY2 ){
		if ( format == RGB24 ){
			mmx_YUY2toRGB24(buffer,out,buffer+width*height*2,width*2,width*2,0);
		} else {
			mmx_YUY2toRGB32(buffer,out,buffer+width*height*2,width*2,width*2,0);
		} 
	}
#endif
 
}

// Decompress a YV12 keyframe

void CodecInst::ArithYV12Decompress(){
	
	unsigned char * dst = out;
	unsigned char * dst2 = buffer;
	if ( format == YUY2 ){
		dst = buffer;
		dst2 = out;
	}

	int size = *(UINT32*)(in+1);
	uncompact_macro(in+9,dst,width*height,width,height,&info_a,YV12);
	uncompact_macro(in+size,dst+width*height,width*height/4,width/2,height/2,&info_b,YV12);
	size = *(UINT32*)(in+5);
	uncompact_macro(in+size,dst+width*height+width*height/4,width*height/4,width/2,height/2,NULL,YV12);
	
	ASM_BlockRestore(dst+width*height+width*height/4,width/2,width*height/4,0);
	wait_for_threads(2);
	
	if ( !multithreading ){
		ASM_BlockRestore(dst,width,width*height,0);
		ASM_BlockRestore(dst+width*height,width/2,width*height/4,0);
	}

	if ( format == YV12 )
		return;


	//upsample if needed
#ifndef X64_BUILD
	if ( SSE2 ){
			isse_yv12_to_yuy2(dst,dst+width*height+width*height/4,dst+width*height,width,width,width/2,dst2,width*2,height);
	} else {
			mmx_yv12_to_yuy2(dst,dst+width*height+width*height/4,dst+width*height,width,width,width/2,dst2,width*2,height);
	}
#else
	isse_yv12_to_yuy2(dst,dst+width*height+width*height/4,dst+width*height,width,width,width/2,dst2,width*2,height);
#endif

	if ( format == YUY2 )
		return;

	// upsample to RGB
#ifndef X64_BUILD
	if ( format == RGB32 ){
		mmx_YUY2toRGB32(dst2,out,dst2+width*height*2,width*2);
	} else {
		mmx_YUY2toRGB24(dst2,out,dst2+width*height*2,width*2);
	}
#else
	if ( format == RGB32 ){
		mmx_YUY2toRGB32(dst2,out,dst2+width*height*2,width*2,width*2,1);
	} else {
		mmx_YUY2toRGB24(dst2,out,dst2+width*height*2,width*2,width*2,1);
	}
#endif
}

#ifndef X64_BUILD
#include "convert_xvid.cpp"
#endif

void CodecInst::ReduceResDecompress(){

	width/=2;
	height/=2;
	int size = *(UINT32*)(in+1);
	unsigned char * dest = (format==YV12)?buffer:out;
	uncompact_macro(in+9,dest,width*height,width,height,&info_a,YV12);
	uncompact_macro(in+size,dest+width*height,width*height/4,width/2,height/2,&info_b,YV12);
	size = *(UINT32*)(in+5);
	uncompact_macro(in+size,dest+width*height+width*height/4,width*height/4,width/2,height/2,NULL,YV12);
	
	ASM_BlockRestore(dest+width*height+width*height/4,width/2,width*height/4,0);
	wait_for_threads(2);
	if ( !multithreading ){
		ASM_BlockRestore(dest,width,width*height,0);
		ASM_BlockRestore(dest+width*height,width/2,width*height/4,0);
	}

	width*=2;
	height*=2;

	//crash();

	unsigned char * source=dest;
	dest=(format==YV12)?out:buffer;

	unsigned char * ysrc = source;
	unsigned char * usrc = ysrc+width*height/4;
	unsigned char * vsrc = usrc+width*height/4/4;
	unsigned char * ydest = dest;
	unsigned char * udest = ydest+width*height;
	unsigned char * vdest = udest+width*height/4;

#ifndef X64_BUILD
	enlarge_res(ysrc,ydest,buffer2,width/2,height/2,SSE2,SSE,MMX);
	enlarge_res(usrc,udest,buffer2,width/4,height/4,SSE2,SSE,MMX);
	enlarge_res(vsrc,vdest,buffer2,width/4,height/4,SSE2,SSE,MMX);
#else
	enlarge_res(ysrc,ydest,buffer2,width/2,height/2,1,0,0);
	enlarge_res(usrc,udest,buffer2,width/4,height/4,1,0,0);
	enlarge_res(vsrc,vdest,buffer2,width/4,height/4,1,0,0);
#endif

	ysrc = ydest;
	usrc = udest;
	vsrc = vdest;

#ifdef X64_BUILD
	if ( format == RGB24) {
		isse_yv12_to_yuy2(ysrc,vsrc,usrc,width,width,width/2,buffer2,width*2,height);
		mmx_YUY2toRGB24(buffer2,out,buffer2+width*height*2,width*2,width*2,1);
	} else if ( format == RGB32 ) {
		isse_yv12_to_yuy2(ysrc,vsrc,usrc,width,width,width/2,buffer2,width*2,height);
		mmx_YUY2toRGB32(buffer2,out,buffer2+width*height*2,width*2,width*2,1);
	} else if ( format == YUY2 ) {
		isse_yv12_to_yuy2(ysrc,vsrc,usrc,width,width,width/2,out,width*2,height);
	}
#else
	if ( format == RGB24) {
		yv12_to_rgb24_mmx(out,width,ysrc,vsrc,usrc,width,width/2,width,-(int)height);
	} else if ( format == RGB32 ) {
		yv12_to_rgb32_mmx(out,width,ysrc,vsrc,usrc,width,width/2,width,-(int)height);
	} else if ( format == YUY2 ) {
		yv12_to_yuyv_mmx(out,width,ysrc,vsrc,usrc,width,width/2,width,height);
	}
#endif
	
}

// Called to decompress a frame, the actual decompression will be
// handed off to other functions based on the frame type.
DWORD CodecInst::Decompress(ICDECOMPRESS* icinfo, DWORD dwSize) {
//	try {

//	crash();

	DWORD return_code=ICERR_OK;
	if ( started != 0x1337 ){
		DecompressBegin(icinfo->lpbiInput,icinfo->lpbiOutput);
	}
	out = (unsigned char *)icinfo->lpOutput;
	in  = (unsigned char *)icinfo->lpInput; 
	icinfo->lpbiOutput->biSizeImage = length;

	// according to the avi specs, the calling application is responsible for handling null frames.
	if ( icinfo->lpbiInput->biSizeImage == 0 ){
		#ifndef LAGARITH_RELEASE
		MessageBox (HWND_DESKTOP, "Received request to decode a null frame", "Error", MB_OK | MB_ICONEXCLAMATION);
		#endif
		return ICERR_OK;
	}

	/*interlaced=(in[0]&INTERLACED_FLAG)>0;
	if ( interlaced ){
		//char msg[]="Interlaced frame detected\n";
		//MessageBox (HWND_DESKTOP, msg, "Error", MB_OK | MB_ICONEXCLAMATION);
		width*=2;
		height/=2;
	}*/

	switch ( in[0] ){
	case ARITH_RGB24:
		ArithDecompress();
		break;
	case BYTEFRAME:
		ByteDecompress();
		break;
	case PIXELFRAME:
		PixelDecompress();
		break;
	case PIXELFRAME_ALPHA:
		AlphaPixelDecompress();
		break;
	case ARITH_ALPHA:
		ArithRGBADecompress();
		break;
	case ARITH_YUY2:
		ArithYUY2Decompress();
		break;
	case OBSOLETE_ARITH:
		ArithDecompress();
		break;
	case ARITH_YV12:
		ArithYV12Decompress();
		break;
	case UNALIGNED_RGB24:
		ArithDecompress();
		break;
	case REDUCED_RES:
		ReduceResDecompress();
		break;
	case UNCOMPRESSED:
		memcpy(out,in+1,length);
		break;
	default:
		#ifndef LAGARITH_RELEASE
		char emsg[128];
		sprintf(emsg,"Unrecognized frame type: %d",in[0]);
		MessageBox (HWND_DESKTOP, emsg, "Error", MB_OK | MB_ICONEXCLAMATION);
		#endif
		return_code=ICERR_ERROR;
		break;
	}
	/*if ( interlaced ){
		width/=2;
		height*=2;
	}*/
	return return_code;
//} catch ( ... ){
//	MessageBox (HWND_DESKTOP, "Exception caught in decompress main", "Error", MB_OK | MB_ICONEXCLAMATION);
//	return ICERR_INTERNAL;
//}
}

//MessageBox (HWND_DESKTOP, msg, "Error", MB_OK | MB_ICONEXCLAMATION);